@@ -14,7 +14,7 @@
   Lesser General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public
-  License along with libcanberra. If not, If not, see
+  License along with libcanberra. If not, see
   <http://www.gnu.org/licenses/>.
 ***/
 
@@ -199,6 +199,10 @@ static int translate_error(int error) {
             return CA_ERROR_OOM;
         case -EBUSY:
             return CA_ERROR_NOTAVAILABLE;
+        case -EINVAL:
+            return CA_ERROR_INVALID;
+        case -ENOSYS:
+            return CA_ERROR_NOTSUPPORTED;
         default:
             if (ca_debug())
                 fprintf(stderr, "Got unhandled error from ALSA: %s\n", snd_strerror(error));
@@ -292,13 +296,13 @@ static void* thread_func(void *userdata) {
         goto finish;
     }
 
-    n_pfd = ret + 1;
+    n_pfd = (nfds_t) ret + 1;
     if (!(pfd = ca_new(struct pollfd, n_pfd))) {
         ret = CA_ERROR_OOM;
         goto finish;
     }
 
-    if ((ret = snd_pcm_poll_descriptors(out->pcm, pfd+1, n_pfd-1)) < 0) {
+    if ((ret = snd_pcm_poll_descriptors(out->pcm, pfd+1, (unsigned) n_pfd-1)) < 0) {
         ret = translate_error(ret);
         goto finish;
     }
@@ -309,6 +313,7 @@ static void* thread_func(void *userdata) {
 
     for (;;) {
         unsigned short revents;
+        snd_pcm_sframes_t sframes;
 
         if (out->dead)
             break;
@@ -322,7 +327,7 @@ static void* thread_func(void *userdata) {
         if (pfd[0].revents)
             break;
 
-        if ((ret = snd_pcm_poll_descriptors_revents(out->pcm, pfd+1, n_pfd-1, &revents)) < 0) {
+        if ((ret = snd_pcm_poll_descriptors_revents(out->pcm, pfd+1, (unsigned) n_pfd-1, &revents)) < 0) {
             ret = translate_error(ret);
             goto finish;
         }
@@ -376,9 +381,9 @@ static void* thread_func(void *userdata) {
             break;
         }
 
-        if ((ret = snd_pcm_writei(out->pcm, d, nbytes/fs)) < 0) {
+        if ((sframes = snd_pcm_writei(out->pcm, d, nbytes/fs)) < 0) {
 
-            if ((ret = snd_pcm_recover(out->pcm, ret, 1)) < 0) {
+            if ((ret = snd_pcm_recover(out->pcm, (int) sframes, 1)) < 0) {
                 ret = translate_error(ret);
                 goto finish;
             }
@@ -386,8 +391,8 @@ static void* thread_func(void *userdata) {
             continue;
         }
 
-        nbytes -= ret*fs;
-        d = (uint8_t*) d + ret*fs;
+        nbytes -= (size_t) sframes*fs;
+        d = (uint8_t*) d + (size_t) sframes*fs;
     }
 
     ret = CA_SUCCESS;
@@ -444,7 +449,7 @@ int driver_play(ca_context *c, uint32_t id, ca_proplist *proplist, ca_finish_cal
         goto finish;
     }
 
-    if ((ret = ca_lookup_sound(&out->file, &p->theme, c->props, proplist)) < 0)
+    if ((ret = ca_lookup_sound(&out->file, NULL, &p->theme, c->props, proplist)) < 0)
         goto finish;
 
     if ((ret = open_alsa(c, out)) < 0)
