@@ -14,7 +14,7 @@
   Lesser General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public
-  License along with libcanberra. If not, If not, see
+  License along with libcanberra. If not, see
   <http://www.gnu.org/licenses/>.
 ***/
 
@@ -30,6 +30,7 @@
 #include "sound-theme-spec.h"
 #include "malloc.h"
 #include "llist.h"
+#include "cache.h"
 
 #define DEFAULT_THEME "freedesktop"
 #define FALLBACK_THEME "freedesktop"
@@ -41,7 +42,8 @@ typedef struct ca_data_dir ca_data_dir;
 struct ca_data_dir {
     CA_LLIST_FIELDS(ca_data_dir);
 
-    char *name;
+    char *theme_name;
+    char *dir_name;
     char *output_profile;
 };
 
@@ -55,7 +57,7 @@ struct ca_theme_data {
     ca_bool_t loaded_fallback_theme;
 };
 
-static int get_data_home(char **e) {
+int ca_get_data_home(char **e) {
     const char *env, *subdir;
     char *r;
     ca_return_val_if_fail(e, CA_ERROR_INVALID);
@@ -90,32 +92,41 @@ static ca_bool_t data_dir_matches(ca_data_dir *d, const char*output_profile) {
     return ca_streq(d->output_profile, output_profile);
 }
 
-static ca_data_dir* find_data_dir(ca_theme_data *t, const char *name) {
+static ca_data_dir* find_data_dir(ca_theme_data *t, const char *theme_name, const char *dir_name) {
     ca_data_dir *d;
 
     ca_assert(t);
-    ca_assert(name);
+    ca_assert(theme_name);
+    ca_assert(dir_name);
 
     for (d = t->data_dirs; d; d = d->next)
-        if (ca_streq(d->name, name))
+        if (ca_streq(d->theme_name, theme_name) &&
+            ca_streq(d->dir_name, dir_name))
             return d;
 
     return NULL;
 }
 
-static int add_data_dir(ca_theme_data *t, const char *name) {
+static int add_data_dir(ca_theme_data *t, const char *theme_name, const char *dir_name) {
     ca_data_dir *d;
 
     ca_return_val_if_fail(t, CA_ERROR_INVALID);
-    ca_return_val_if_fail(name, CA_ERROR_INVALID);
+    ca_return_val_if_fail(theme_name, CA_ERROR_INVALID);
+    ca_return_val_if_fail(dir_name, CA_ERROR_INVALID);
 
-    if (find_data_dir(t, name))
+    if (find_data_dir(t, theme_name, dir_name))
         return CA_SUCCESS;
 
     if (!(d = ca_new0(ca_data_dir, 1)))
         return CA_ERROR_OOM;
 
-    if (!(d->name = ca_strdup(name))) {
+    if (!(d->theme_name = ca_strdup(theme_name))) {
+        ca_free(d);
+        return CA_ERROR_OOM;
+    }
+
+    if (!(d->dir_name = ca_strdup(dir_name))) {
+        ca_free(d->theme_name);
         ca_free(d);
         return CA_ERROR_OOM;
     }
@@ -185,7 +196,7 @@ static int load_theme_path(ca_theme_data *t, const char *prefix, const char *nam
                 goto fail;
             }
 
-            current_data_dir = find_data_dir(t, d);
+            current_data_dir = find_data_dir(t, name, d);
             ca_free(d);
 
             in_sound_theme_section = FALSE;
@@ -227,7 +238,7 @@ static int load_theme_path(ca_theme_data *t, const char *prefix, const char *nam
                             goto fail;
                         }
 
-                        ret = add_data_dir(t, p);
+                        ret = add_data_dir(t, name, p);
                         ca_free(p);
 
                         if (ret != CA_SUCCESS)
@@ -248,13 +259,15 @@ static int load_theme_path(ca_theme_data *t, const char *prefix, const char *nam
 
             if (!strncmp(ln, "OutputProfile=", 14)) {
 
-                if (current_data_dir->output_profile && !ca_streq(current_data_dir->output_profile, ln+14)) {
-                    ret = CA_ERROR_CORRUPT;
-                    goto fail;
-                }
+                if (!current_data_dir->output_profile) {
 
-                if (!(current_data_dir->output_profile = ca_strdup(ln+14))) {
-                    ret = CA_ERROR_OOM;
+                    if (!(current_data_dir->output_profile = ca_strdup(ln+14))) {
+                        ret = CA_ERROR_OOM;
+                        goto fail;
+                    }
+
+                } else if (!ca_streq(current_data_dir->output_profile, ln+14)) {
+                    ret = CA_ERROR_CORRUPT;
                     goto fail;
                 }
 
@@ -302,6 +315,15 @@ fail:
     return ret;
 }
 
+const char *ca_get_data_dirs(void) {
+    const char *g;
+
+    if (!(g = getenv("XDG_DATA_DIRS")) || *g == 0)
+        return "/usr/local/share:/usr/share";
+
+    return g;
+}
+
 static int load_theme_dir(ca_theme_data *t, const char *name) {
     int ret;
     char *e;
@@ -311,12 +333,12 @@ static int load_theme_dir(ca_theme_data *t, const char *name) {
     ca_return_val_if_fail(name, CA_ERROR_INVALID);
     ca_return_val_if_fail(t->n_theme_dir < N_THEME_DIR_MAX, CA_ERROR_CORRUPT);
 
-    if ((ret = get_data_home(&e)) < 0)
-        return ret;
-
     if (ca_streq(name, FALLBACK_THEME))
         t->loaded_fallback_theme = TRUE;
 
+    if ((ret = ca_get_data_home(&e)) < 0)
+        return ret;
+
     if (e) {
         ret = load_theme_path(t, e, name);
         ca_free(e);
@@ -325,8 +347,7 @@ static int load_theme_dir(ca_theme_data *t, const char *name) {
             return ret;
     }
 
-    if (!(g = getenv("XDG_DATA_DIRS")) || *g == 0)
-        g = "/usr/local/share:/usr/share";
+    g = ca_get_data_dirs();
 
     for (;;) {
         size_t k;
@@ -398,7 +419,9 @@ fail:
 
 static int find_sound_for_suffix(
         ca_sound_file **f,
-        ca_theme_data *t,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        const char *theme_name,
         const char *name,
         const char *path,
         const char *suffix,
@@ -409,14 +432,15 @@ static int find_sound_for_suffix(
     int ret;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name, CA_ERROR_INVALID);
     ca_return_val_if_fail(path, CA_ERROR_INVALID);
     ca_return_val_if_fail(path[0] == '/', CA_ERROR_INVALID);
 
     if (!(fn = ca_sprintf_malloc("%s%s%s%s%s%s%s/%s%s",
                                  path,
-                                 t ? "/" : "",
-                                 t ? t->name : "",
+                                 theme_name ? "/" : "",
+                                 theme_name ? theme_name : "",
                                  subdir ? "/" : "",
                                  subdir ? subdir : "",
                                  locale ? "/" : "",
@@ -432,16 +456,21 @@ static int find_sound_for_suffix(
             ret = errno == ENOENT ? CA_ERROR_NOTFOUND : CA_ERROR_SYSTEM;
 
     } else
-        ret = ca_sound_file_open(f, fn);
+        ret = sfopen(f, fn);
 
-    ca_free(fn);
+    if (ret == CA_SUCCESS && sound_path)
+        *sound_path = fn;
+    else
+        ca_free(fn);
 
     return ret;
 }
 
 static int find_sound_in_locale(
         ca_sound_file **f,
-        ca_theme_data *t,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        const char *theme_name,
         const char *name,
         const char *path,
         const char *locale,
@@ -451,6 +480,7 @@ static int find_sound_in_locale(
     char *p;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name && *name, CA_ERROR_INVALID);
     ca_return_val_if_fail(path, CA_ERROR_INVALID);
     ca_return_val_if_fail(path[0] == '/', CA_ERROR_INVALID);
@@ -460,9 +490,10 @@ static int find_sound_in_locale(
 
     sprintf(p, "%s/sounds", path);
 
-    if ((ret = find_sound_for_suffix(f, t, name, p, ".disabled", locale, subdir)) == CA_ERROR_NOTFOUND)
-        if ((ret = find_sound_for_suffix(f, t, name, p, ".ogg", locale, subdir)) == CA_ERROR_NOTFOUND)
-            ret = find_sound_for_suffix(f, t, name, p, ".wav", locale, subdir);
+    if ((ret = find_sound_for_suffix(f, sfopen, sound_path, theme_name, name, p, ".disabled", locale, subdir)) == CA_ERROR_NOTFOUND)
+        if ((ret = find_sound_for_suffix(f, sfopen, sound_path,theme_name, name, p, ".oga", locale, subdir)) == CA_ERROR_NOTFOUND)
+            if ((ret = find_sound_for_suffix(f, sfopen, sound_path,theme_name, name, p, ".ogg", locale, subdir)) == CA_ERROR_NOTFOUND)
+                ret = find_sound_for_suffix(f, sfopen, sound_path,theme_name, name, p, ".wav", locale, subdir);
 
     ca_free(p);
 
@@ -471,7 +502,9 @@ static int find_sound_in_locale(
 
 static int find_sound_for_locale(
         ca_sound_file **f,
-        ca_theme_data *theme,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        const char *theme_name,
         const char *name,
         const char *path,
         const char *locale,
@@ -481,22 +514,23 @@ static int find_sound_for_locale(
     int ret;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name && *name, CA_ERROR_INVALID);
     ca_return_val_if_fail(path, CA_ERROR_INVALID);
     ca_return_val_if_fail(locale, CA_ERROR_INVALID);
 
     /* First, try the locale def itself */
-    if ((ret = find_sound_in_locale(f, theme, name, path, locale, subdir)) != CA_ERROR_NOTFOUND)
+    if ((ret = find_sound_in_locale(f, sfopen, sound_path, theme_name, name, path, locale, subdir)) != CA_ERROR_NOTFOUND)
         return ret;
 
     /* Then, try to truncate at the @ */
     if ((e = strchr(locale, '@'))) {
         char *t;
 
-        if (!(t = ca_strndup(locale, e - locale)))
+        if (!(t = ca_strndup(locale, (size_t) (e - locale))))
             return CA_ERROR_OOM;
 
-        ret = find_sound_in_locale(f, theme, name, path, t, subdir);
+        ret = find_sound_in_locale(f, sfopen, sound_path, theme_name, name, path, t, subdir);
         ca_free(t);
 
         if (ret != CA_ERROR_NOTFOUND)
@@ -507,10 +541,10 @@ static int find_sound_for_locale(
     if ((e = strchr(locale, '_'))) {
         char *t;
 
-        if (!(t = ca_strndup(locale, e - locale)))
+        if (!(t = ca_strndup(locale, (size_t) (e - locale))))
             return CA_ERROR_OOM;
 
-        ret = find_sound_in_locale(f, theme, name, path, t, subdir);
+        ret = find_sound_in_locale(f, sfopen, sound_path, theme_name, name, path, t, subdir);
         ca_free(t);
 
         if (ret != CA_ERROR_NOTFOUND)
@@ -519,16 +553,18 @@ static int find_sound_for_locale(
 
     /* Then, try "C" as fallback locale */
     if (strcmp(locale, "C"))
-        if ((ret = find_sound_in_locale(f, theme, name, path, "C", subdir)) != CA_ERROR_NOTFOUND)
+        if ((ret = find_sound_in_locale(f, sfopen, sound_path, theme_name, name, path, "C", subdir)) != CA_ERROR_NOTFOUND)
             return ret;
 
     /* Try without locale */
-    return find_sound_in_locale(f, theme, name, path, NULL, subdir);
+    return find_sound_in_locale(f, sfopen, sound_path, theme_name, name, path, NULL, subdir);
 }
 
 static int find_sound_for_name(
         ca_sound_file **f,
-        ca_theme_data *t,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        const char *theme_name,
         const char *name,
         const char *path,
         const char *locale,
@@ -538,9 +574,10 @@ static int find_sound_for_name(
     const char *k;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name && *name, CA_ERROR_INVALID);
 
-    if ((ret = find_sound_for_locale(f, t, name, path, locale, subdir)) != CA_ERROR_NOTFOUND)
+    if ((ret = find_sound_for_locale(f, sfopen, sound_path, theme_name, name, path, locale, subdir)) != CA_ERROR_NOTFOUND)
         return ret;
 
     k = strchr(name, 0);
@@ -555,10 +592,10 @@ static int find_sound_for_name(
 
         } while (*k != '-');
 
-        if (!(n = ca_strndup(name, k-name)))
+        if (!(n = ca_strndup(name, (size_t) (k-name))))
             return CA_ERROR_OOM;
 
-        if ((ret = find_sound_for_locale(f, t, n, path, locale, subdir)) != CA_ERROR_NOTFOUND) {
+        if ((ret = find_sound_for_locale(f, sfopen, sound_path, theme_name, n, path, locale, subdir)) != CA_ERROR_NOTFOUND) {
             ca_free(n);
             return ret;
         }
@@ -569,7 +606,9 @@ static int find_sound_for_name(
 
 static int find_sound_in_subdir(
         ca_sound_file **f,
-        ca_theme_data *t,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        const char *theme_name,
         const char *name,
         const char *locale,
         const char *subdir) {
@@ -579,21 +618,21 @@ static int find_sound_in_subdir(
     const char *g;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name, CA_ERROR_INVALID);
 
-    if ((ret = get_data_home(&e)) < 0)
+    if ((ret = ca_get_data_home(&e)) < 0)
         return ret;
 
     if (e) {
-        ret = find_sound_for_name(f, t, name, e, locale, subdir);
+        ret = find_sound_for_name(f, sfopen, sound_path, theme_name, name, e, locale, subdir);
         ca_free(e);
 
         if (ret != CA_ERROR_NOTFOUND)
             return ret;
     }
 
-    if (!(g = getenv("XDG_DATA_DIRS")) || *g == 0)
-        g = "/usr/local/share:/usr/share";
+    g = ca_get_data_dirs();
 
     for (;;) {
         size_t k;
@@ -606,7 +645,7 @@ static int find_sound_in_subdir(
             if (!(p = ca_strndup(g, k)))
                 return CA_ERROR_OOM;
 
-            ret = find_sound_for_name(f, t, name, p, locale, subdir);
+            ret = find_sound_for_name(f, sfopen, sound_path, theme_name, name, p, locale, subdir);
             ca_free(p);
 
             if (ret != CA_ERROR_NOTFOUND)
@@ -624,6 +663,8 @@ static int find_sound_in_subdir(
 
 static int find_sound_in_profile(
         ca_sound_file **f,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
         ca_theme_data *t,
         const char *name,
         const char *locale,
@@ -633,13 +674,14 @@ static int find_sound_in_profile(
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
     ca_return_val_if_fail(t, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(name, CA_ERROR_INVALID);
 
     for (d = t->data_dirs; d; d = d->next)
         if (data_dir_matches(d, profile)) {
             int ret;
 
-            if ((ret = find_sound_in_subdir(f, t, name, locale, d->name)) != CA_ERROR_NOTFOUND)
+            if ((ret = find_sound_in_subdir(f, sfopen, sound_path, d->theme_name, name, locale, d->dir_name)) != CA_ERROR_NOTFOUND)
                 return ret;
         }
 
@@ -648,6 +690,8 @@ static int find_sound_in_profile(
 
 static int find_sound_in_theme(
         ca_sound_file **f,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
         ca_theme_data *t,
         const char *name,
         const char *locale,
@@ -657,25 +701,28 @@ static int find_sound_in_theme(
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
     ca_return_val_if_fail(name, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(profile, CA_ERROR_INVALID);
 
     if (t) {
         /* First, try the profile def itself */
-        if ((ret = find_sound_in_profile(f, t, name, locale, profile)) != CA_ERROR_NOTFOUND)
+        if ((ret = find_sound_in_profile(f, sfopen, sound_path, t, name, locale, profile)) != CA_ERROR_NOTFOUND)
             return ret;
 
         /* Then, fall back to stereo */
         if (!ca_streq(profile, DEFAULT_OUTPUT_PROFILE))
-            if ((ret = find_sound_in_profile(f, t, name, locale, DEFAULT_OUTPUT_PROFILE)) != CA_ERROR_NOTFOUND)
+            if ((ret = find_sound_in_profile(f, sfopen, sound_path, t, name, locale, DEFAULT_OUTPUT_PROFILE)) != CA_ERROR_NOTFOUND)
                 return ret;
     }
 
     /* And fall back to no profile */
-    return find_sound_in_subdir(f, t, name, locale, NULL);
+    return find_sound_in_subdir(f, sfopen, sound_path, t ? t->name : NULL, name, locale, NULL);
 }
 
 static int find_sound_for_theme(
         ca_sound_file **f,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
         ca_theme_data **t,
         const char *theme,
         const char *name,
@@ -686,6 +733,7 @@ static int find_sound_for_theme(
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
     ca_return_val_if_fail(t, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
     ca_return_val_if_fail(theme, CA_ERROR_INVALID);
     ca_return_val_if_fail(name && *name, CA_ERROR_INVALID);
     ca_return_val_if_fail(locale, CA_ERROR_INVALID);
@@ -697,21 +745,33 @@ static int find_sound_for_theme(
             ret = load_theme_data(t, FALLBACK_THEME);
 
     if (ret == CA_SUCCESS)
-        if ((ret = find_sound_in_theme(f, *t, name, locale, profile)) != CA_ERROR_NOTFOUND)
+        if ((ret = find_sound_in_theme(f, sfopen, sound_path, *t, name, locale, profile)) != CA_ERROR_NOTFOUND)
             return ret;
 
     /* Then, fall back to "unthemed" files */
-    return find_sound_in_theme(f, NULL, name, locale, profile);
+    return find_sound_in_theme(f, sfopen, sound_path, NULL, name, locale, profile);
 }
 
-int ca_lookup_sound(ca_sound_file **f, ca_theme_data **t, ca_proplist *cp, ca_proplist *sp) {
-    int ret;
+int ca_lookup_sound_with_callback(
+        ca_sound_file **f,
+        ca_sound_file_open_callback_t sfopen,
+        char **sound_path,
+        ca_theme_data **t,
+        ca_proplist *cp,
+        ca_proplist *sp) {
+    int ret = CA_ERROR_INVALID;
     const char *name, *fname;
 
     ca_return_val_if_fail(f, CA_ERROR_INVALID);
     ca_return_val_if_fail(t, CA_ERROR_INVALID);
     ca_return_val_if_fail(cp, CA_ERROR_INVALID);
     ca_return_val_if_fail(sp, CA_ERROR_INVALID);
+    ca_return_val_if_fail(sfopen, CA_ERROR_INVALID);
+
+    *f = NULL;
+
+    if (sound_path)
+        *sound_path = NULL;
 
     ca_mutex_lock(cp->mutex);
     ca_mutex_lock(sp->mutex);
@@ -734,12 +794,42 @@ int ca_lookup_sound(ca_sound_file **f, ca_theme_data **t, ca_proplist *cp, ca_pr
             if (!(profile = ca_proplist_gets_unlocked(cp, CA_PROP_CANBERRA_XDG_THEME_OUTPUT_PROFILE)))
                 profile = DEFAULT_OUTPUT_PROFILE;
 
-        ret = find_sound_for_theme(f, t, theme, name, locale, profile);
+#ifdef HAVE_CACHE
+        if ((ret = ca_cache_lookup_sound(f, sfopen, sound_path, theme, name, locale, profile)) >= 0) {
 
-    } else if ((fname = ca_proplist_gets_unlocked(sp, CA_PROP_MEDIA_FILENAME)))
-        ret = ca_sound_file_open(f, fname);
-    else
-        ret = CA_ERROR_INVALID;
+            /* This entry is available in the cache, let's transform
+             * negative cache entries to CA_ERROR_NOTFOUND */
+
+            if (!*f)
+                ret = CA_ERROR_NOTFOUND;
+
+        } else {
+            char *spath = NULL;
+
+            /* Either this entry was not available in the database,
+             * neither positive nor negative, or the database was
+             * corrupt, or it was out-of-date. In all cases try to
+             * find the entry manually. */
+
+            if ((ret = find_sound_for_theme(f, sfopen, sound_path ? sound_path : &spath, t, theme, name, locale, profile)) >= 0)
+                /* Ok, we found it. Let's update the cache */
+                ca_cache_store_sound(theme, name, locale, profile, sound_path ? *sound_path : spath);
+            else if (ret == CA_ERROR_NOTFOUND)
+                /* Doesn't seem to be around, let's create a negative cache entry */
+                ca_cache_store_sound(theme, name, locale, profile, NULL);
+
+            ca_free(spath);
+        }
+
+#else
+        ret = find_sound_for_theme(f, sfopen, sound_path, t, theme, name, locale, profile);
+#endif
+    }
+
+    if (ret == CA_ERROR_NOTFOUND || !name) {
+        if ((fname = ca_proplist_gets_unlocked(sp, CA_PROP_MEDIA_FILENAME)))
+            ret = sfopen(f, fname);
+    }
 
     ca_mutex_unlock(cp->mutex);
     ca_mutex_unlock(sp->mutex);
@@ -747,6 +837,16 @@ int ca_lookup_sound(ca_sound_file **f, ca_theme_data **t, ca_proplist *cp, ca_pr
     return ret;
 }
 
+int ca_lookup_sound(
+        ca_sound_file **f,
+        char **sound_path,
+        ca_theme_data **t,
+        ca_proplist *cp,
+        ca_proplist *sp) {
+
+    return ca_lookup_sound_with_callback(f, ca_sound_file_open, sound_path, t, cp, sp);
+}
+
 void ca_theme_data_free(ca_theme_data *t) {
     ca_assert(t);
 
@@ -755,7 +855,8 @@ void ca_theme_data_free(ca_theme_data *t) {
 
         CA_LLIST_REMOVE(ca_data_dir, t->data_dirs, d);
 
-        ca_free(d->name);
+        ca_free(d->theme_name);
+        ca_free(d->dir_name);
         ca_free(d->output_profile);
         ca_free(d);
     }
